@@ -1,6 +1,6 @@
 /* tsinh -- test file for mpc_sinh.
 
-Copyright (C) 2008 INRIA
+Copyright (C) 2008, 2013 INRIA
 
 This file is part of GNU MPC.
 
@@ -20,15 +20,22 @@ along with this program. If not, see http://www.gnu.org/licenses/ .
 
 #include "mpc-tests.h"
 
+#define MPC_FUNCTION_CALL                                       \
+  P[0].mpc_inex = mpc_sinh (P[1].mpc, P[2].mpc, P[3].mpc_rnd)
+#define MPC_FUNCTION_CALL_REUSE_OP1                             \
+  P[0].mpc_inex = mpc_sinh (P[1].mpc, P[1].mpc, P[3].mpc_rnd)
+
+#include "data_check.tpl"
+#include "tgeneric.tpl"
+
 int
 main (void)
 {
-  DECL_FUNC (CC, f, mpc_sinh);
-
   test_start ();
 
-  data_check (f, "sinh.dat");
-  tgeneric (f, 2, 512, 7, 7);
+  data_check_template ("sinh.dsc", "sinh.dat");
+
+  tgeneric_template ("sinh.dsc", 2, 512, 7, 7);
 
   test_end ();
 
